
from collections import OrderedDict, MutableMapping

class LruDict(MutableMapping):
    def __init__(self, maxlen, items=None):
        self._maxlen = maxlen
        self.d = OrderedDict()
        if items:
            for k, v in items:
                self[k] = v

    @property
    def maxlen(self):
        return self._maxlen

    def __getitem__(self, key):
        self.d.move_to_end(key)
        return self.d[key]

    def __setitem__(self, key, value):
        if key in self.d:
            self.d.move_to_end(key)
        elif len(self.d) == self.maxlen:
            self.d.popitem(last=False)
        self.d[key] = value

    def __delitem__(self, key):
        del self.d[key]

    def __iter__(self):
        return self.d.__iter__()

    def __len__(self):
        return len(self.d)




from .statserver import setup_statserver, stats_timer
from .internals import Protocol
from .internals import WorkServer as CWorkServer
from .workserver import *
from .lrudict import LruDict

__version__=0.3




from sanic import Sanic
from sanic import response
from sanic.exceptions import NotFound
import asyncio
import pygal
import psutil

import mrworkserver
from mrhttp import escape
import tenjin, os
from tenjin.helpers import *
tenjin.set_template_encoding('utf-8')
engine = tenjin.Engine(path=[os.path.dirname(__file__)+'/html'],escapefunc='escape',tostrfunc='str')

from pygal.style import DarkStyle
cstyle = DarkStyle(
  title_font_size=28,
  label_font_size=24,
  major_label_font_size=24
)

def setup_statserver(ws):

  app = Sanic(__name__,configure_logging=False)
  @app.route("/<name>")
  async def test(r,name):
    if name == "time":
      data = ws.async_times
      title = 'Processing time (ms)'
    elif name == "cpu":
      data = ws.cpu
      title = 'CPU Utilization'
    elif name == "mem":
      data = ws.mem
      title = 'Memory Usage'
    elif name in ws.counts:
      data = ws.counts[name]["cnts"]
      title = ws.counts[name]["title"]
    else:
      raise NotFound("Page not found")
    b = await ws.loop.run_in_executor( ws.procpool, blocks, title, data )
    return response.raw(b,headers={'Content-Type': 'image/svg+xml'})

  @app.route("/")
  async def index(r):
    imgs = ["time","cpu","mem"]
    imgs.extend( ws.counts.keys() )
    return response.html(engine.render('stats.ten', {"imgs":imgs}))

  server = app.create_server(host="0.0.0.0", port=5000)
  return asyncio.ensure_future(server)

async def stats_timer(self):
  minute = 0
  while 1:
    await asyncio.sleep(60)
    self.mem.append( psutil.virtual_memory().percent )
    self.cpu.append( psutil.cpu_percent() )

    for k in self.counts:
      c = self.counts[k]
      c["cnts"].append( c["cnt"] )
      c["cnt"] = 0

    if len( self.async_times ) > 0:
      tot = 0
      num = 0
      for ms in self.async_times:
        tot += ms
        num += 1
      avg = tot/num
      self.async_times_1m.append(avg)
      #self.async_times = []
    minute += 1
    if minute == 1440:
      # TODO Store 10m times to disk
      pass


def blocks(title, data):
  try:
    line_chart = pygal.Line(style=cstyle,show_legend=False)
    line_chart.title = title
    line_chart.add('', data)
    return line_chart.render()
  except Exception as e:
    print("chart exception", args, str(e))
    return b''





import socket, os, asyncio, time, struct
from concurrent.futures import ProcessPoolExecutor
from io import BytesIO

from mrworkserver import Protocol, CWorkServer
from mrworkserver import setup_statserver, stats_timer


class WorkServer(CWorkServer):
  def __init__(self, *, protocol_factory=None, seconds_to_gather=0, collect_stats=False, callback=None, fetch_callback=None):
    self._loop = asyncio.new_event_loop()
    asyncio.set_event_loop(self.loop)
    self._connections = set()
    self._protocol_factory = protocol_factory or Protocol
    self._seconds_to_gather = seconds_to_gather
    self.cb = callback
    self.fetchcb = fetch_callback
    self.async_times = None
    self.collect_stats = False
    self.async_times_1m = []
    self.stats_task = None
    self.webserver_task = None
    if collect_stats:
      self.collect_stats = True
      self.cdata = {}
      
      self.async_times = []
      self.mem = []
      self.cpu = []
      self.counts = {}
      self.procpool = ProcessPoolExecutor(2)


      self.webserver_task = setup_statserver(self)

      #app = Sanic(__name__,configure_logging=False)
      #@app.route("/<name>")
      #async def test(r,name):
        #if name == "time": 
          #data = self.async_times
          #title = 'Processing time (ms)'
        #elif name == "cpu": 
          #data = self.cpu
          #title = 'CPU Utilization'
        #elif name == "mem": 
          #data = self.mem
          #title = 'Memory Usage'
        #elif name in self.counts:
          #data = self.counts[name]["cnts"]
          #title = self.counts[name]["title"]
        #else:
          #raise NotFound("OK")
        #b = await self.loop.run_in_executor( self.procpool, blocks, title, data )
        #return response.raw(b,headers={'Content-Type': 'image/svg+xml'})
#
      #server = app.create_server(host="0.0.0.0", port=5000)
      #self.webserver_task = asyncio.ensure_future(server)



    self.on_start = None
    self.on_stop  = None
    #self.workq = asyncio.Queue(loop=self.loop)
    super(WorkServer,self).__init__(callback, seconds_to_gather)


  @property
  def loop(self):
    if not self._loop:
      self._loop = asyncio.new_event_loop()
    return self._loop

  #def dcConnect(servers):
    #for s in servers:
      #srv = Server( s[0], s[1] )
      #srv.r, srv.w = await asyncio.open_connection( s[0], s[1], loop=self._loop, limit=DEFAULT_BUFFER_SIZE, ssl=ssl)
      #self.servers.append(srv)

  # Fetch
  def prefetch(self, o):
    b = self.fetchcb(self, o)
    if not isinstance(b, bytes): return ""
    b = b'\x02' + struct.pack("<I", len(b)) + b
    return b    


  # Stats
  def count_title(self, name, title):
    if not name in self.counts:
      self.counts[name] = { "title":title, "cnt":0, "cnts":[] }
    self.counts[name]["title"] = title

  def count(self, name, cnt):
    if not name in self.counts:
      self.counts[name] = { "title":name, "cnt":0, "cnts":[] }
    self.counts[name]["cnt"] += cnt 


  def run(self, host='0.0.0.0', port=7100, *, cores=None, ssl=None):

    if not asyncio.iscoroutinefunction(self.cb):
      print("WorkServer.cb must be an async function")
      return;

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    os.set_inheritable(sock.fileno(), True)

    loop = self.loop
    #asyncio.set_event_loop(loop)
    if self.on_start:
      loop.run_until_complete( self.on_start(self) )

    if self.collect_stats: 
      self.stats_task = asyncio.ensure_future( stats_timer(self) )

    server_coro = loop.create_server( lambda: self._protocol_factory(self), sock=sock, ssl=ssl)
    server = loop.run_until_complete(server_coro)

    try:
      loop.run_forever()
    except KeyboardInterrupt:
      pass
    finally:
      if self.stats_task: self.stats_task.cancel()
      if self.webserver_task: self.webserver_task.cancel()
      server.close()
      loop = asyncio.get_event_loop()
      if self.on_stop:
        loop.run_until_complete( self.on_stop(self) )
      loop.run_until_complete(server.wait_closed())
      loop.close()

#class ChartData:
  #title = ""
  #data = []
  

def blocks(title, data):
  try:
    line_chart = pygal.Line()
    line_chart.title = title
    line_chart.add('avg', data)
    return line_chart.render()
  except Exception as e:
    print("chart exception", args, str(e))
    return b''
    

  #b = BytesIO()
  #plt.plot(data)
  #plt.savefig(b, format='png')
  #return b.getvalue()
  #plt.savefig('foo.png', bbox_inches='tight') 






import codecs
from setuptools import setup, Extension, find_packages

#from distutils.command.install import INSTALL_SCHEMES
#for scheme in INSTALL_SCHEMES.values():
    #scheme['data'] = scheme['purelib']

m1 = Extension(
    'mrworkserver.internals',
     sources = [ 
      './src/mrworkserver/internals/dec.c',
      './src/mrworkserver/internals/module.c',
      './src/mrworkserver/internals/protocol.c',
      './src/mrworkserver/internals/ws.c',
      './src/mrworkserver/internals/unpack.c',
     ],
     include_dirs = ['./src/mrworkserver/internals'],
     extra_compile_args = ['-msse4.2', '-mavx2', '-mbmi2', '-Wunused-variable','-std=c99','-Wno-discarded-qualifiers', '-Wno-unused-variable','-Wno-unused-function'],
     extra_link_args = [],
     #extra_link_args = ['-lasan'],
     define_macros = [('DEBUG_PRINT',1)]
)

with codecs.open('README.md', encoding='utf-8') as f:
    README = f.read()


setup(
  name="mrworkserver", 
  version="0.7",
  license='MIT',
  description='A python work server written in C',
  long_description = README,
  long_description_content_type='text/markdown',
  ext_modules = [m1],
  package_dir={'':'src'},
  packages=find_packages('src'),# + ['prof'],
  #data_files=[('', ['html'])],
  include_package_data=True,
  #package_data={'prof': ['prof.so']},
  install_requires=[
    #'uvloop<0.9.0',
    'uvloop>0.9.0',
  ],
  platforms='x86_64 Linux and MacOS X',
  url='http://github.com/MarkReedZ/mrworkserver/',
  author='Mark Reed',
  author_email='markreed99@gmail.com',
  keywords=['web', 'asyncio'],
  classifiers=[
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Environment :: Web Environment',
    'License :: OSI Approved :: MIT License',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: C',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Internet :: WWW/HTTP'
   ]
)

